# AUTOGENERATED! DO NOT EDIT! File to edit: 03_01_trajectories.ipynb (unless otherwise specified).

__all__ = ['plot_trajectory', 'tortuosity', 'map_size', 'avg_distance_between_pair', 'random_color', 'draw_pair_points',
           'plot_all_env']

# Cell
import open3d as o3d
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import pylab as plot
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import yaml
import math
import os
import random
from queue import Queue, PriorityQueue
plt.rcParams["figure.figsize"] = (16, 16)

# Cell
def plot_trajectory(data_root, env, quantity=None):
    """
        Calculate metrics for the input map.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
            quantity(int) : The draw quantity of trajectories. Default is None means total.

        Returns:
            A figure (trajectory.png) would be saved.
    """
    # Default path & file name
    folder_map = 'npy' #The relative path to find the yaml file `obstacle_map.map`.
    map_file = 'obstacle_map.npy' #The file name of the 2D occupency grid map.
    folder_path = 'trajectory' #The relative path to find the yaml file of trajectories.
    folder_info = 'environment_info' #The relative path to find the yaml file of information about 2D occupency grid map.
    info_file = 'grid_map_info.yaml' #The file name of the yaml file about 2D occupency grid map.

    yaml_path = data_root + '/' + env + '/' + folder_info + '/' + info_file
    with open(yaml_path, 'r') as f:
            info_yaml_data = yaml.load(f, Loader=yaml.CLoader)

    map_path = data_root + '/' + env + '/' + folder_map + '/' + map_file
    vis_trajectory_map = np.load(map_path)

    trajectory_folder = data_root + '/' + env + '/' + folder_path
    trajectory_yaml_list = os.listdir(trajectory_folder)

    color = ['white', 'black', 'red', 'green', 'orange', 'purple','gray', 'brown', 'yellow', 'blue']
    used_color = color[0:len(trajectory_yaml_list)+2]
    cmap = ListedColormap(used_color)
    legend = []
    for index,baseline in enumerate(trajectory_yaml_list):
        legend.append(mpatches.Patch(color=used_color[index+2], label=baseline))
        baseline_path = trajectory_folder + '/' + baseline
        with open(baseline_path, 'r') as stream:
            yaml_data = yaml.load(stream, Loader=yaml.CLoader)

        if(quantity == None):
            for path in yaml_data["trajectories"]:
                for pose in path:
                    c = int((pose["position"][0] - info_yaml_data["origin_x"])/info_yaml_data["resolution"])
                    r = int((pose["position"][1] - info_yaml_data["origin_y"])/info_yaml_data["resolution"])

                    if c >= info_yaml_data["map_width"] or c < 0 : continue
                    if r >= info_yaml_data["map_height"] or r < 0 : continue
                    vis_trajectory_map[r][c] = index + 2
        else:
            for i in range(int(quantity)):
                path = yaml_data["trajectories"][i]
                for pose in path:
                    c = int((pose["position"][0] - info_yaml_data["origin_x"])/info_yaml_data["resolution"])
                    r = int((pose["position"][1] - info_yaml_data["origin_y"])/info_yaml_data["resolution"])

                    if c >= info_yaml_data["map_width"] or c < 0 : continue
                    if r >= info_yaml_data["map_height"] or r < 0 : continue
                    vis_trajectory_map[r][c] = index + 2

    fig, ax = plt.subplots()
    ax.imshow(vis_trajectory_map)
    ax.legend(handles=legend, prop={'size': 16})
    barrr = ax.imshow(vis_trajectory_map,cmap=cmap)

    figure_path = data_root + '/' + env + '/' + 'figure'
    if not os.path.exists(figure_path):
        os.makedirs(os.path.join(data_root + '/' + env, 'figure'))

    fig.savefig(figure_path + '/trajectory.png')

# Cell
def tortuosity(yaml_path):
    """
        Tortuosity calculated using the formula (arc length / chord length)
        Args:
            yaml_path (String) : The full path of yaml file.

        Returns:
            avg_tortuosity (Float) : The average tortuosity of all pathes in such yaml file.
    """
    dist_between_points = lambda x1, y1, x2, y2: ((x2-x1)**2 + (y2-y1)**2)**0.5

#     print(yaml_path)
    with open(yaml_path, 'r') as stream:
        yaml_data = yaml.load(stream, Loader=yaml.CLoader)

    total = 0.0
    for i_path, path in enumerate(yaml_data["trajectories"]):
        arc_len = 0.0
        for i_pose, pose in enumerate(path):
            if i_pose == 0: continue
            x1, y1 = path[i_pose-1]["position"][0], path[i_pose-1]["position"][1]
            x2, y2 = path[i_pose]["position"][0], path[i_pose]["position"][1]
            arc_len += dist_between_points(x1, y1, x2, y2)

        x_first, y_first = path[0]["position"][0], path[0]["position"][1]
        x_last, y_last = path[-1]["position"][0], path[-1]["position"][1]
        chord_len = dist_between_points(x_first, y_first, x_last, y_last)
        tor = arc_len / chord_len
#         print("The " ,i_path+1,"th path with tortuosity " ,tor)
        total += tor
    avg = total/len(yaml_data["trajectories"])
#     print("Avg tortuosity of all path: ", avg)
    return avg

# Cell
def map_size(data_root, env):
    """
        Real size of the environment.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
        Returns:
            map_width (Int) : Map width (meter).
            map_height (Int) : Map width (meter).
    """
    yaml_path = data_root + '/' + env + '/' + 'environment_info' + '/' + 'grid_map_info.yaml'
    with open(yaml_path, 'r') as f:
        info_yaml_data = yaml.load(f, Loader=yaml.CLoader)
    map_width = info_yaml_data["map_width"]*info_yaml_data["resolution"]
    map_height = info_yaml_data["map_height"]*info_yaml_data["resolution"]
    print("Map Width: ",map_width)
    print("Map Height: ",map_height)
    return map_width, map_height

def avg_distance_between_pair(data_root, env, file_name=None):
    """
        Calculate average distance between start point and goal point.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
            file_name (String) : The file stored all pairs.
        Returns:
            avg_dis (Float) : The average of the distance between Start and Goal points (meter).
    """
    dist_between_points = lambda x1, y1, x2, y2: ((x2-x1)**2 + (y2-y1)**2)**0.5

    if(file_name):
        pair_yaml_path = data_root + '/' + env + '/' + 'environment_info' + '/' + file_name +'.yaml'
    else:
        pair_yaml_path = data_root + '/' + env + '/' + 'environment_info' + '/' + env +'.yaml'
    with open(pair_yaml_path, 'r') as stream:
        yaml_data = yaml.load(stream, Loader=yaml.CLoader)

    total_dis = 0
    for pair in yaml_data["pairs"]:
        x1, y1 = pair["start"][0], pair["start"][1]
        x2, y2 = pair["goal"][0], pair["goal"][1]
        dis = dist_between_points(x1, y1, x2, y2)
        total_dis += dis

    avg_distance = total_dis/len(yaml_data["pairs"])
    print("The average of the distance between Start and Goal points: ", avg_distance)
    return avg_distance


# Cell
def random_color(number):
    '''
    generate random colors for each pair points and the number of color should not less than point's number.
    Args :
        number (int) : total number points.

    Returns :
        unique_color (list) : a list of different color.
    '''
    color=['#FFFFFF','#000000']
    if len(color)<number:
        for i in range (number):
            color.append("#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)]))
            i+=1
    unique_color = []
    for name in color:         # 1st loop
        if name not in unique_color:   # 2nd loop
            unique_color.append(name)
    if (len(unique_color)<number+2):
        unique_color = random_color(number)
    return unique_color

def draw_pair_points(data_root, env, file_name=None):
    '''
    Draw every pair points in the map from path : 'env'/environment_info/'env'.yaml .
    Args :
        data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
        env (String) : Which environment was used.
        file_name (String) : The file stored all pairs.
    Returns:
        A figure (pairPoint.png) would be saved.
    '''
    # default path
    folder_map = 'npy' # The relative path to find the yaml file `bit.map`.
    map_file = 'obstacle_map.npy' # The file name of the 2D occupency grid map.
    info_file = 'grid_map_info.yaml' #The file name of the yaml file about 2D occupency grid map.

    # grid_map_info
    yaml_path = data_root + '/' + env + '/' + 'environment_info/' + info_file
    with open(yaml_path, 'r') as f:
        info_yaml_data = yaml.load(f, Loader=yaml.CLoader)
    # obstacle map
    map_path = data_root + '/' + '/' + env + '/npy/' + map_file
    pair_points_map = np.load(map_path)

    # start-goal pairs
    if(file_name):
        pair_yaml_path = data_root + '/' + env + '/' + 'environment_info' + '/' + file_name+'.yaml'
    else:
        pair_yaml_path = data_root + '/' + env + '/' + 'environment_info' + '/' + env+'.yaml'
    with open(pair_yaml_path, 'r') as stream:
        yaml_data = yaml.load(stream, Loader=yaml.CLoader)

    # color
    color=random_color(len(yaml_data["pairs"]))
    cmap=ListedColormap(color)
    #color_index change .npy map's color.
    color_index = 2

    for pair in yaml_data["pairs"]:
        x1, y1 = pair["start"][0], pair["start"][1]
        x2, y2 = pair["goal"][0], pair["goal"][1]

        c1 = int((x1 - info_yaml_data["origin_x"]) / info_yaml_data["resolution"])
        c2 = int((x2 - info_yaml_data["origin_x"]) / info_yaml_data["resolution"])
        r1 = int((y1 - info_yaml_data["origin_y"]) / info_yaml_data["resolution"])
        r2 = int((y2 - info_yaml_data["origin_y"]) / info_yaml_data["resolution"])

        # Line from interpolation
        delta_c = (c1-c2)*0.001
        delta_r = (r1-r2)*0.001
        for delta in range(1000):
            c = int( c2 + delta * delta_c )
            r = int( r2 + delta * delta_r )
            pair_points_map[r][c] = color_index
#         pair_points_map[r2][c2] = color_index

        color_index += 1

    fig, ax = plt.subplots()
    ax.set_title(f'{env} pair points map',color='r')
    ax.imshow(pair_points_map)
    barrr = ax.imshow(pair_points_map,cmap=cmap)

    figure_path = data_root + '/' + env + '/' + 'figure'
    if not os.path.exists(figure_path):
        os.makedirs(os.path.join( data_root + '/' + env, 'figure'))

    fig.savefig(figure_path + '/pairPoint.png')


# Cell
def plot_all_env(data_root, which_metric):
    '''
    Use sub-figure to visualize the metric of all environments together.
    You can plot 1.pairPoint 2.trajectory.
    you must run draw_pair_points() and plot_trajectory() to get all env's pairPoint.png and trajectory.png !
    Args:
        data_root (String) : Path. All directory under this path is an environment. Please check file structure in 03_diff_visualization.ipynb
        which_metric (String) : Metric name.
    Returns:
        None. A figure will pop up.

    '''
    envs = os.listdir(data_root)
    envs.remove('stage1')
    envs.remove('cave')
    img = []
    for i, env in enumerate(envs):
        img.append(Image.open(data_root+'/'+env+'/figure/'+which_metric+'.png'))
    fig, ((ax_1, ax_2, ax_3), (ax_4, ax_5, ax_6))= plt.subplots(nrows=2, ncols=3)
    fig.subplots_adjust(left=0,
                        bottom=0.1,
                        right=7,
                        top=5,
                        wspace=0,
                        hspace=0)
    ax_1.imshow(img[0])
    ax_1.axis("off")
    ax_2.imshow(img[1])
    ax_2.axis("off")
    ax_3.imshow(img[2])
    ax_3.axis("off")
    ax_4.imshow(img[3])
    ax_4.axis("off")
    ax_5.imshow(img[4])
    ax_5.axis("off")
    ax_6.imshow(img[5])
    ax_6.axis("off")

    plt.show()


# AUTOGENERATED! DO NOT EDIT! File to edit: 10_gym_environment.ipynb (unless otherwise specified).

__all__ = ['read_path', 'show_env']

# Cell
import os
import re

def read_path(path, output_name):
    """

    This function will read path of gym information, and output a txt file that get all gym environment.

    Args:
        path : The gym init file path
        output_name : The name of output file
    Return:
        file_path : The stored all gym environment file path

    """
    file_path = os.path.join("./Data/" + output_name)

    os.system("cat " + path + " | grep \"id=\" >> " + file_path)

    return file_path

def show_env(path):
    """

    This function will read gym environment file and show all gym.

    Args :
        path : The gym environment file path
    Return :
        link : list all gym environment.
    """

    env = []
    f = open(path, 'r')
    for line in f.readlines():
      if(line.find('#')):
        flag = [_.start() for _ in re.finditer("\'",line)]
        env.append(line[flag[0]+1:flag[1]])
    f.close()

    return env


import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from copy import copy
from torch.nn.parameter import Parameter
from torchvision import models

# visionnet_input = True
# logits_input = False
# knob_pos_smoothening = False
class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out

class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size(0), -1)

class ReduceMean(nn.Module):
    def forward(self, x):
        return torch.mean(x, dim=1, keepdim=True)

class SpatialSoftmax(nn.Module):
    def __init__(self, height, width, channel, temperature=None, data_format='NCHW'):
        super(SpatialSoftmax, self).__init__()
        self.data_format = data_format
        self.height = height
        self.width = width
        self.channel = channel

        if temperature:
            self.temperature = Parameter(torch.ones(1)*temperature)
        else:
            self.temperature = 1.

        pos_x, pos_y = torch.meshgrid(torch.linspace(-1., 1., self.height),torch.linspace(-1., 1., self.width))

        pos_x = pos_x.reshape(self.height*self.width).float()
        pos_y = pos_y.reshape(self.height*self.width).float()

        self.register_buffer('pos_x', pos_x)
        self.register_buffer('pos_y', pos_y)

    def forward(self, feature):
        # Output:
        #   (N, C*2) x_0 y_0 ...
        if self.data_format == 'NHWC':
            feature = feature.transpose(1, 3).tranpose(2, 3).view(-1, self.height*self.width)
        else:
            feature = feature.view(-1, self.height*self.width)
        softmax_attention = F.softmax(feature/self.temperature, dim=-1)
        expected_x = torch.sum(self.pos_x*softmax_attention, dim=1, keepdim=True)
        expected_y = torch.sum(self.pos_y*softmax_attention, dim=1, keepdim=True)
        expected_xy = torch.cat([expected_x, expected_y], 1)
        feature_keypoints = expected_xy.view(-1, self.channel*2)
        return feature_keypoints


class VisionModel(nn.Module):
    def __init__(self, cam="front"):
        super().__init__()

        self.front_origin = torch.tensor([[-0.5, 1.0]]).cuda() #y,z
        self.top_origin = torch.tensor([[0.5, 0.0]]).cuda() #x,y

        # input is nx3x256x256
        self.backbone = nn.Sequential(
            BasicBlock(3, 64, stride=2),
            BasicBlock(64, 64, stride=2),
            BasicBlock(64, 64, stride=1),
        )

        self.rm = nn.Sequential(
            ReduceMean()
        )

        self.position_regressor = nn.Sequential(
            ReduceMean(),
            Flatten(),
            nn.Linear(4096, 32),
            nn.Tanh(),
            nn.Linear(32, 2),
        )

    def forward(self, x):
        backbone = self.backbone(x)
        heatmap_features = self.rm(backbone)        
        pos_regression = self.position_regressor(backbone)
        return pos_regression, heatmap_features

class VisionModelXYZ(nn.Module):
    def __init__(self, cam="xyz"):
        super().__init__()
        self.top_net = VisionModel()
        self.front_net = VisionModel()

    def forward(self, x, y):
        pp1, hm1 = self.top_net.forward(x)
        pp2, hm2 = self.front_net.forward(y)
        pp = torch.stack((pp1[:,0], (pp1[:,1]+pp2[:,0])/2., pp2[:,1]), 1)
        return pp, hm1, hm2

if __name__ == "__main__":
    vision = VisionModel()
    input_data = torch.zeros(5, 3, 256, 256)
    input_data[0, 0, 32, 32] = 1
    input_data[1, 0, 32, 32] = 1
    input_data[2, 0, 32, 32] = 1
    input_data[3, 0, 32, 32] = 1
    input_data[4, 0, 32, 32] = 1
    # print(vision.pos_from_logits(logit, cam="front"))

    model_parameters = filter(lambda p: p.requires_grad, vision.parameters())
    params = sum([np.prod(p.size()) for p in model_parameters])
    print("Model has {} parameters...".format(params))

    logits, features = vision(input_data)

    # print(logits.size(), features.size())

        # def global2label_fr(self, x):
    #     x = x - self.front_origin.data.cpu().numpy()
    #     x[:,0], x[:,1] = -x[:,1], x[:,0].copy()
    #     x = (x + 0.5) * 64
    #     x = np.int8(np.round(x))
    #     label = self.gkern(64, 64, (x[0,1],x[0,0]),s=1)
    #     return label

    # def global2label_top(self, x):
    #     x = x - self.top_origin.data.cpu().numpy()
    #     x[:,0], x[:,1] = -x[:,1], x[:,0].copy()
    #     x = (x + 0.5) * 64
    #     x = np.int8(np.round(x))
    #     label = self.gkern(64, 64, (x[0,1],x[0,0]),s=1)
    #     # label = np.zeros((64,64))
    #     # label[x[0,0],x[0,1]] = 10
    #     return label



# AUTOGENERATED! DO NOT EDIT! File to edit: 03_diff_visualization.ipynb (unless otherwise specified).

__all__ = ['DifficultyMetrics', 'read_ply', 'calculate_metric', 'plot_metric']

# Cell
import open3d as o3d
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import pylab as plot
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import yaml
import math
import os
import random
from queue import Queue, PriorityQueue
plt.rcParams["figure.figsize"] = (16, 16)

# Cell
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import pylab as plot
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import yaml
import math
import os
import random
from queue import Queue, PriorityQueue
plt.rcParams["figure.figsize"] = (16, 16)

#export
class DifficultyMetrics:
    def __init__(self): pass

    def crop_sub_map(self,r,c,full_occu_map):
        """
        Crop a sub-map with fixed size (31*31) from the full occupancy grid map.
        The pixel full_occu_map[r][c] will be the center of sub-map, that is crop_map[15][15].
        Padding with 1, which means consider as obstacle if a pixel we need in sub-map in out of the full occupancy grid map.

        Args:
            full_occu_map (2D array) : The full occupancy grid map.
            r(int) : index in row in the full occupancy grid map. .
            c(int) : index in column in the full occupancy grid map. .
        Returns:
            crop_map (2D array) : An 31*31 sub-map from full_occu_map.
        """
        crop_map = np.ones((31,31))
        for sub_r in range(-15,16):
            for sub_c in range(-15,16):
                if (r+sub_r < 0) or (r+sub_r >=full_occu_map.shape[0]): continue
                if (c+sub_c < 0) or (c+sub_c >=full_occu_map.shape[1]): continue
                crop_map[sub_r+15][sub_c+15] = full_occu_map[r+sub_r][c+sub_c]
        return crop_map

    def closest_wall(self, full_occu_map):
        """
        Calculate the avg Distance to Closest Obstacle of an occupancy grid map.
        Args:
            full_occu_map (2D array) : The full occupancy grid map
        Returns:
            Avg Distance to Closest Obstacle of all cell in map.
        """
        dists = [[0 for i in range(len(full_occu_map[0]))] for j in range(len(full_occu_map))]
        for r in range(len(full_occu_map)):
            for c in range(len(full_occu_map[0])):
                crop_map = self.crop_sub_map(r, c, full_occu_map)
                dists[r][c] = self._dist_closest_wall(crop_map)
        return dists

    def avg_visibility(self, full_occu_map):
        """
        Calculate the average visibility of an occupancy grid map.
        Args:
            full_occu_map (2D array) : The full occupancy grid map
        Returns:
            The average visibility at each point.
        """
        vis = [[0 for i in range(len(full_occu_map[0]))] for j in range(len(full_occu_map))]
        for r in range(len(full_occu_map)):
            for c in range(len(full_occu_map[0])):
                crop_map = self.crop_sub_map(r, c, full_occu_map)
                vis[r][c] = self._avg_vis_pixel(crop_map)
        return vis

    def dispersion(self, full_occu_map):
        """
        Calculate the average dispersion of an occupancy grid map.
        Checks along 16 axes within the dispersion radius.
        Args:
            full_occu_map (2D array) : The full occupancy grid map.
        Returns:
            The average dispersion at each point.
        """
        disp = [[0 for i in range(len(full_occu_map[0]))] for j in range(len(full_occu_map))]
        for r in range(len(full_occu_map)):
            for c in range(len(full_occu_map[0])):
                crop_map = self.crop_sub_map(r, c, full_occu_map)
                disp[r][c] = self._cell_dispersion(crop_map)
        return disp

    def characteristic_dimension(self,full_occu_map):
        """
            Calculate the average dispersion of an occupancy grid map.
            Characteristic dimension calculated in 2 directions for 4 axes
            Args:
                full_occu_map (2D array) : The full occupancy grid map.
            Returns:
                The average of characteristic dimension at each point.
            """

        cdr = [[0 for i in range(len(full_occu_map[0]))] for j in range(len(full_occu_map))]
        for r in range(len(full_occu_map)):
            for c in range(len(full_occu_map[0])):
                if full_occu_map[r][c] == 1:
                    cdr[r][c] = 0
                    continue
                crop_map = self.crop_sub_map(r, c, full_occu_map)
                cdr_min = ((15*2 + 15*2) ** 0.5) * 2 # diagonal
                for axis in [(0, 1), (1, 1), (1, 0), (1, -1)]:
                    cdr_min = min(cdr_min, self._distance(axis, crop_map))
                cdr[r][c] = cdr_min
        return cdr

    def _distance(self, axis, crop_map):
        """
        Args:
            crop_map (2D array) : occupancy grid map.
            r(int) : index in row.
            c(int) : index in column.
            axis(int) : a list such as [0,1], indicate the directions.
        Returns:
            The distance along the axis in both directions, not including (r, c)
        """
        r = 15
        c = 15
        if crop_map[r][c] == 1: return -1# is obstalce

        # check axis in both directions
        reverse_axis = (axis[0] * -1, axis[1] * -1)
        dist = 0
        for move in [axis, reverse_axis]:
            r_curr = r
            c_curr = c
            # move in axis direction until an obstacle is found
            while crop_map[r_curr][c_curr] != 1:
                r_curr += move[0]
                c_curr += move[1]

                # if not in map
                if (r_curr+move[0]>=31 or c_curr+move[1]>=31 or r_curr+move[0]<0 or c_curr+move[1]<0 ): break
                # add the distance traveled to the total
                if crop_map[r_curr][c_curr] != 1:
                    dist += math.sqrt(move[0] ** 2 + move[1] ** 2)
        return dist


    def _dist_between_points(self, a, b):
        """
        Args:
            a and b are tuples : (row, col)
        Returns:
            The distance between points
        """
        return math.sqrt((a[0] - b[0]) ** 2 + (a[1] - b[1]) ** 2)

    def _cell_dispersion(self, crop_map):
        """
            Args:
                crop_map (2D array) : occupancy grid map.
                r(int) : index in row.
                c(int) : index in column.
                radius(int) : cell size, default as 1.
            Returns:
                The cell dispersion within the radius.
            """
        radius=1
        r = 15
        c = 15
        if crop_map[r][c] == 1: return -1
        axes_wall = []
        # four cardinal, four diagonal, and one in between each (slope +- 1/2 or +-2)
        for move in [(0, 1), (1, 2), (1, 1), (2, 1), (1, 0), (2, -1), (1, -1), (1, -2), (0, -1), (-2, -1), (-1, -1), (-1, -2), (-1, 0), (-2, 1), (-1, 1), (-1, 2)]:
            count = 0
            wall = False
            r_curr = r
            c_curr = c
            while count < radius and not wall:
                r_curr += move[0]
                c_curr += move[1]
                if r_curr < 0 or r_curr >= len(crop_map) or c_curr < 0 or c_curr >= len(crop_map[0]):break
                if crop_map[r_curr][c_curr] == 1:wall = True
                if move[0] == 2 or move[1] == 2:
                    count += 2 # count the in-between axes as two steps
                else: count += 1

            if wall: axes_wall.append(True)
            else: axes_wall.append(False)
            # count the number of changes in this cell's field of view
        change_count = 0
        for i in range(len(axes_wall)-1):
            if axes_wall[i] != axes_wall[i+1]:
                change_count += 1
            if axes_wall[0] != axes_wall[15]:
                change_count += 1
        return change_count

    def _avg_vis_pixel(self, crop_map):
        """
            Args:
                crop_map (2D array) : occupancy grid map.
                r(int) : index in row.
                c(int) : index in column.
            Returns:
                The average visibility at the point (r, c).
            """
        r = 15
        c = 15
        total_vis = 0.0
        num_axes = 0
        for r_move in [-1, 0, 1]:
            for c_move in [-1, 0, 1]:
                if r_move == 0 and c_move == 0:continue
                this_vis = 0
                r_curr = r
                c_curr = c
                wall_found = False
                while not wall_found:
                    if ( r_curr>=31 or c_curr>=31 or r_curr<0 or c_curr<0 ) : break # not in map
                    if crop_map[r_curr][c_curr] == 1: wall_found = True
                    else: this_vis += math.sqrt((r_move ** 2) + (c_move ** 2))
                    r_curr += r_move
                    c_curr += c_move
                total_vis += this_vis
                num_axes += 1
        return total_vis / num_axes

    def _in_map(self, r, c, full_occu_map):
        """
        Bounds check.
        Args:
            r(int) : index in row.
            c(int) : index in column.
        Returns:
            A Boolean means that if the cell is in bounded.
        """
        return r >= 0 and r < len(full_occu_map) and c >= 0 and c < len(full_occu_map[0])


    def _dist_closest_wall(self, crop_map):
        """
            Args:
                crop_map (2D array) : occupancy grid map.
                r(int) : index in row.
                c(int) : index in column.
            Returns:
                The distance to the closest obstacle at point (r, c).
                Returns 0 if crop_map[r][c] is an obstacle, 1 if an adjacent non-diagonal cell is an obstacle, etc.
            """
        pq = PriorityQueue()
        r = 15
        c = 15
        first_wrapper = self.Wrapper(0, r, c)
        pq.put(first_wrapper)
        visited = {(r, c) : first_wrapper}
        while not pq.empty():
            point = pq.get()
            if crop_map[point.r][point.c] == 1: return point.dist # found an obstacle!
            else:
                # enqueue all neighbors if they are in the map and have not been visited
                for row in range(point.r - 1, point.r + 2):
                    for col in range(point.c - 1, point.c + 2):
                        if (row<31 and row>=0 and col<31 and col>=0) and (row, col) not in visited:
                            dist = math.sqrt((row - r) ** 2 + (col - c) ** 2)
                            neighbor = self.Wrapper(dist, row, col)
                            pq.put(neighbor)
                            visited[(row, col)] = neighbor
        # in case the queue is empty before a wall is found (shouldn't happen),
        # the farthest a cell can be from a wall is half the board, since the top and bottom rows are all walls
        return (len(crop_map) - 1) / 2

    class Wrapper:
        """ wrapper class for coordinates """
        def __init__(self, distance, row, col):
            self.dist = distance
            self.r = row
            self.c = col
        def __lt__(self, value): return self.dist < value.dist


# Cell
def read_ply(data_root, env, folder, filename):
    """
        Generate the 2D occupency grid map by 3D pointcloud map.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
            folder (String) : The relative path to find the yaml file.
            filename (String) : The file name of yaml.

            Finally, the full path of the target file would be "data_root + '/' + env + '/' + folder + '/' + filename".
        Returns:
            map_bit (array) : The occupency grid map in the form of an numpy array. It would also be saved in .npy file and visualized in .png file.
    """
    # read ply file
    yaml_path = data_root + '/' + env + '/' + folder + '/' + filename
    print(yaml_path)
    with open(yaml_path, 'r') as stream:
        yaml_data = yaml.load(stream, Loader=yaml.CLoader)
    pc = o3d.io.read_point_cloud(data_root + '/' + env + '/' + folder + '/' + yaml_data['source'])

    points_array = np.asarray((pc.points))

    map_height, map_width = yaml_data['map_height'], yaml_data['map_width']
    map = np.zeros((map_height, map_width))
    origin_x, origin_y = yaml_data['origin_x'], yaml_data['origin_y']
    resolution = yaml_data['resolution']
    for i in range(len(points_array)):
        if points_array[i][2] < yaml_data['z_max'] : continue
        if points_array[i][2] > yaml_data['z_min'] : continue

        c = int((points_array[i][0] - origin_x)/resolution)
        r = int((points_array[i][1] - origin_y)/resolution)

        if c >= map_width or c < 0 :
            continue
        if r >= map_height or r < 0 :
            continue

        map[r][c] += 1

    map_bit = np.zeros(map.shape)
    for c in range(map_width):
        for r in range(map_height):
            map_bit[r][c] = 1 if map[r][c] > yaml_data['threshold'] else 0

    folder_path = os.path.join(data_root + '/' + env, 'npy')
    if not os.path.exists(folder_path):
        os.makedirs(os.path.join(data_root + '/' + env, 'npy'))

    npy_file_name = data_root + '/' + env + '/npy/' + 'obstacle_map.npy'
    np.save(npy_file_name, map_bit)
    plt.imshow(map_bit)
    plt.colorbar()
    plt.show()
    return map_bit


# Cell

def calculate_metric(data_root, env, folder, filename):
    """
        Calculate metrics for the input map.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Which environment was used.
            folder (String) : The relative path to find the yaml file.
            filename (String) : The file name of yaml.

            Finally, the full path of the .npy file for the occupency grid map would be "data_root + '/' + env + '/' + folder + '/' + filename".

        Returns:
            metric_list (array) : The result of metrics which order in [`obstacle map`,`closest wall`,`visibility`, `dispersion`,`characteristic dimension`].

    """
    map_path = data_root + '/' + env + '/' + folder + '/' + filename
    load_obstacle_map = np.load(map_path)
    shape = load_obstacle_map.shape
    diff = DifficultyMetrics()
    total_obstacle_map = load_obstacle_map
    total_closest_wall = diff.closest_wall(load_obstacle_map)
    total_avg_visibility = diff.avg_visibility(load_obstacle_map)
    total_dispersion = diff.dispersion(load_obstacle_map)
    total_characteristic_dimension = diff.characteristic_dimension(load_obstacle_map)

    str_list = ["obstacle_map",\
                "closest_wall",\
                "avg_visibility",\
                "dispersion",\
                "characteristic_dimension"]

    folder_path = os.path.join(data_root + '/' + env, 'npy')
    if not os.path.exists(folder_path):
        os.makedirs(os.path.join(data_root + '/' + env, 'npy'))

    for i,item in enumerate([total_obstacle_map,total_closest_wall,total_avg_visibility, \
                             total_dispersion,total_characteristic_dimension]):
        npy_file_name = data_root + '/' + env + '/npy/' + str_list[i]+".npy"
        np.save(npy_file_name, np.asarray(item))
    metric_list = [total_obstacle_map, total_closest_wall, total_avg_visibility, total_dispersion, total_characteristic_dimension]
    return metric_list

def plot_metric(data_root, env, metric_list):
    """
        Calculate metrics for the input map.
        Args:
            data_root (String) : The path of folder env_exp_files. Default at 'Data/env_exp_files' .
            env (String) : Name of the environment.
            metric_list (array) : The result of metrics which order in [`obstacle map`,`closest wall`,`visibility`, `dispersion`,`characteristic dimension`].

        Returns:
            None : Metrics visualization be saved in .png files. Full path is    data_root + '/' + env + '/figure/' + "XXX.png"
    """

    str_list = ["obstacle_map",\
                "closest_wall",\
                "avg_visibility",\
                "dispersion",\
                "characteristic_dimension"]
    cmap_list = ["binary","RdYlGn","RdYlGn","RdYlGn","Greys"]

    figure_path = os.path.join(data_root + '/' + env, 'figure')
    if not os.path.exists(figure_path):
        os.makedirs(os.path.join(data_root + '/' + env, 'figure'))


    for i,item in enumerate(metric_list):
        fig, ax = plt.subplots()
        barrr = ax.imshow(item,cmap=cmap_list[i], interpolation='nearest')
        ax.set_title(str_list[i])
        fig.colorbar(barrr, ax=ax)

        png_name = data_root + '/' + env + '/figure/' + str_list[i]+".png"
        fig.savefig(png_name)




# AUTOGENERATED! DO NOT EDIT! File to edit: 15_Train_env_analyze.ipynb (unless otherwise specified).

__all__ = ['get_environments_metrics_traj', 'get_metric_values', 'Normalization_1D', 'visualize_1D', 'visualize_2D']

# Cell
import open3d as o3d
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import pylab as plot
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import yaml
import math
import os
from queue import Queue, PriorityQueue
import numpy as np
from sklearn.datasets import load_iris
import matplotlib.pyplot as plt
from sklearn.manifold import MDS
from sklearn.preprocessing import MinMaxScaler
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
from mpl_toolkits.axes_grid1 import make_axes_locatable
from sklearn.preprocessing import normalize
plt.rcParams["figure.figsize"] = (16, 16)

# Cell
def get_environments_metrics_traj(data_root):
    """
    Calculate the average metric of all environment.
    Args:
        data_root (String) : Path. All directory under this path is an environment. Please check file structure in 03_diff_visualization.ipynb
    Returns:
        all_env (Dict) : A Dictionary like this
                            { 'garage': { 'avg_visibility': 15.987035875225928,
                                          'characteristic_dimension': 6.787662532508829,
                                          'closest_wall': 2.254822777080131,
                                          'dispersion': 1.2634011090573012},
                              'forest': { 'avg_visibility': 33.32919485818832,
                                          'characteristic_dimension': 26.731711305286364,
                                          'closest_wall': 3.6290804054099532,
                                          'dispersion': 0.6094594594594595}, ...}
    """
    envs = os.listdir(data_root)
    metrics = ['avg_visibility','characteristic_dimension','closest_wall','dispersion']

    all_env = dict()
    for env in envs:
        each_env = dict()
        # grid_map_info
        yaml_path = data_root + '/' + env + '/' + 'environment_info/' + 'grid_map_info.yaml'
        with open(yaml_path, 'r') as f:
            info_yaml_data = yaml.load(f, Loader=yaml.CLoader)

        # trajectory
        pair_yaml_path = data_root + '/' + env + '/' + 'trajectory' + '/' + env+'_TARE.yaml'
        with open(pair_yaml_path, 'r') as stream:
            pair_yaml_data = yaml.load(stream, Loader=yaml.CLoader)

        column_row = []
        for trajectory in pair_yaml_data["trajectories"]:
            for pose in trajectory:
                x = pose["position"][0]
                y = pose["position"][1]
                c = int((x - info_yaml_data["origin_x"]) / info_yaml_data["resolution"])
                r = int((y - info_yaml_data["origin_y"]) / info_yaml_data["resolution"])
                if [c, r] not in column_row: column_row.append( [c, r] )

        for metric in metrics:
            metric_data = np.load(data_root + '/' + env + '/' + 'npy' + '/' + metric+'.npy')
            norm_metric_data = metric_data

            total = 0.0
            for cr in column_row:
                if cr[1]>=norm_metric_data.shape[0] or cr[0]>=norm_metric_data.shape[1]: continue
                total += norm_metric_data[cr[1]][cr[0]]
            avg = total/len(column_row)

            each_env[metric] = avg
        each_env["tortuosity"] =  tortuosity('Data/env_exp_files/'+env+'/trajectory/'+env+'_TARE.yaml')
        all_env[env] = each_env
    return all_env

# Cell
def get_metric_values(all_env):
    """
    Calculate MAX, MIN, mean & std. for all metrics.
    Args
        all_env (Dict) : A Dictionary like this
                            { 'garage': { 'avg_visibility': 15.987035875225928,
                                          'characteristic_dimension': 6.787662532508829,
                                          'closest_wall': 2.254822777080131,
                                          'dispersion': 1.2634011090573012,
                                          'tortuosity': 1.291879835480693},
                              'forest': { 'avg_visibility': 33.32919485818832,
                                          'characteristic_dimension': 26.731711305286364,
                                          'closest_wall': 3.6290804054099532,
                                          'dispersion': 0.6094594594594595,
                                          'tortuosity': 1.291879835480693}, ...}
    Returns:
        None. But printtn the results.
    """
    metrics = ['avg_visibility', 'characteristic_dimension', 'closest_wall', 'dispersion', 'tortuosity']
    all_data = [[] for i in range(len(metrics))]
    for env_key in all_env:
        for i,metric in enumerate(metrics):
            all_data[i].append( np.array(all_env[env_key][metric]) )
    for i in range(len(all_data)):
        print(metrics[i])
        print('min=', round(np.min(all_data[i]),2),end=' ')
        print('max=', round(np.max(all_data[i]),2),end=' ')
        print('mean=', round(np.mean(all_data[i]),2),end=' ')
        print('std=', round(np.std(all_data[i]),2),end=' \n')

# Cell
def Normalization_1D(arr):
    """
    Normalization. Make the MAX in the array to be 1 and the MIN to be 0.
    Args
        arr (array) : 1D array.
    Returns:
        norm_arr (array) : 1D array.
    """
    MAX = np.max(arr)
    MIN = np.min(arr)

    norm_arr = []
    for i in range(len(arr)):
        norm_i = float( (arr[i]-MIN) / (MAX - MIN) )
        norm_arr.append(norm_i)
    return np.array( norm_arr )

def visualize_1D(all_env):
    """
    Visualize 1D distribution of metrics. All value been normalized to 0~1.
    Args
        all_env (Dict) : A Dictionary like this
                            { 'garage': { 'avg_visibility': 15.987035875225928,
                                          'characteristic_dimension': 6.787662532508829,
                                          'closest_wall': 2.254822777080131,
                                          'dispersion': 1.2634011090573012,
                                          'tortuosity': 1.291879835480693},
                              'forest': { 'avg_visibility': 33.32919485818832,
                                          'characteristic_dimension': 26.731711305286364,
                                          'closest_wall': 3.6290804054099532,
                                          'dispersion': 0.6094594594594595,
                                          'tortuosity': 1.291879835480693}, ...}
    Returns:
        None. But plot the figure.
    """
    metrics = ['avg_visibility', 'characteristic_dimension', 'closest_wall', 'dispersion', 'tortuosity']
    colors = ['red','green','yellow','grey','purple','orange','pink','cyan','silver']
    env_names = []
    all_data = [[] for i in range(len(metrics))]
    for env_key in all_env:
        env_names.append(env_key)
        for i,metric in enumerate(metrics):
            all_data[i].append( np.array(all_env[env_key][metric]) )
    all_data = [all_data[i] for i in range(len(all_data))]

    norm_all_data = [ Normalization_1D(all_data[i]) for i in range(len(all_data)) ]
    plt.figure()
    for i in range(len(norm_all_data)):
        plt.subplot(str(510+i))
        plt.title(metrics[i])
        for j,env_name in enumerate(env_names):
            plt.scatter([norm_all_data[i][j]],[0],color=colors[j],s=300 )

    if not os.path.exists('Data'):
        os.makedirs(os.path.join('Data'))
    plt.savefig("Data/diff_metrics_visualize_1D.png")
    print(env_names)
    print(colors)


# Cell
def visualize_2D(all_env):
    """
    Visualize 2D distribution of metrics.
    Consider all combinations of two out of five metrics. Therefore, here are 10 figures.
    All value been normalized to 0~1.
    Args
        all_env (Dict) : A Dictionary like this
                            { 'garage': { 'avg_visibility': 15.987035875225928,
                                          'characteristic_dimension': 6.787662532508829,
                                          'closest_wall': 2.254822777080131,
                                          'dispersion': 1.2634011090573012,
                                          'tortuosity': 1.291879835480693},
                              'forest': { 'avg_visibility': 33.32919485818832,
                                          'characteristic_dimension': 26.731711305286364,
                                          'closest_wall': 3.6290804054099532,
                                          'dispersion': 0.6094594594594595,
                                          'tortuosity': 1.291879835480693}, ...}
    Returns:
        None. But plot the figures.
    """
    metrics = ['avg_visibility', 'characteristic_dimension', 'closest_wall', 'dispersion', 'tortuosity']
    env_names = []
    colors = []
    all_data = [[] for i in range(len(metrics))]
    for env_key in all_env:
        env_names.append(env_key)
        if env_key=='cave' or env_key=='BARN_3x3':colors.append('green')
        else:colors.append('red')

        for i,metric in enumerate(metrics):
            all_data[i].append( np.array(all_env[env_key][metric]) )
    all_data = [all_data[i] for i in range(len(all_data))]
    norm_all_data = [ Normalization_1D(all_data[i]) for i in range(len(all_data)) ]

    if not os.path.exists('Data'): os.makedirs(os.path.join('Data'))
    plt.figure()
    for i in range(5):
        for j in range(i+1,5):
            fig, ax = plt.subplots()
            ax.scatter(norm_all_data[i],norm_all_data[j],s=300,color=colors)
            for k, font in enumerate (env_names):
                plt.annotate(font, (norm_all_data[i][k],norm_all_data[j][k]),size=20)
            ax.set_xlabel(metrics[i])
            ax.set_ylabel(metrics[j])
            ax.grid()
            fig.savefig('Data/'+metrics[i]+'_'+metrics[j]+'.png')
